#include "ehs/Log.h"
#include "ehs/HRNG.h"

#include <sys/syscall.h>
#include <sys/mman.h>
#include <unistd.h>
#include <cstdarg>

extern "C"
{
	long syscall(long number, ...) {
		va_list args;
		va_start(args, number);

		// Retrieve the system call arguments.
		unsigned long arg1 = va_arg(args, unsigned long);
		unsigned long arg2 = va_arg(args, unsigned long);
		unsigned long arg3 = va_arg(args, unsigned long);
		unsigned long arg4 = va_arg(args, unsigned long);
		unsigned long arg5 = va_arg(args, unsigned long);
		unsigned long arg6 = va_arg(args, unsigned long);
		va_end(args);

		// Variable to store the system call return value
		long ret;

		// Inline assembly to perform the system call
		__asm__ volatile (
				"mov %1, %%rax\n\t" // System call number -> RAX
				"mov %2, %%rdi\n\t" // First argument -> RDI
				"mov %3, %%rsi\n\t" // Second argument -> RSI
				"mov %4, %%rdx\n\t" // Third argument -> RDX
				"mov %5, %%r10\n\t" // Fourth argument -> R10
				"mov %6, %%r8\n\t"  // Fifth argument -> R8
				"mov %7, %%r9\n\t"  // Sixth argument -> R9
				"syscall\n\t"       // Invoke system call
				"mov %%rax, %0\n\t" // Store return value
				: "=r"(ret) // output
				: "r"(number), "r"(arg1), "r"(arg2), "r"(arg3), "r"(arg4), "r"(arg5), "r"(arg6) // inputs
				: "%rax", "%rdi", "%rsi", "%rdx", "%r10", "%r8", "%r9", "memory", "cc" // clobbers
				);

		return ret;
	}

	int* __errno_location()
	{
		static __thread int tls_errno = 0; // __thread or thread_local (C++11 and later)
		return &tls_errno;
	}

	#ifdef EHS_DEBUG
		unsigned long __stack_chk_guard = 0x0000DEAD;
	#else
		unsigned long __stack_chk_guard = ehs::HRNG::GenerateSeed_u64();
	#endif

	void __stack_chk_fail(void)
	{
		EHS_LOG("Error", 0, "Found corrupted stack.");

		syscall(SYS_exit, 1);
	}

	int __cxa_atexit(void (*func)(void *), void *arg, void * /*dso_handle*/)
	{
		return 0;
	}

	size_t strlen(const char *s)
	{
		size_t count = 0;
		while (s[count])
			count++;

		return count;
	}
}

void* operator new[](unsigned long size) noexcept(false)
{
	syscall(SYS_brk,)
	void* ptr = (void*)syscall(SYS_mmap, nullptr, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
	if (ptr == MAP_FAILED)
	{
		EHS_LOG("Error", 0, "Failed to allocate memory on the heap.");
		return nullptr;
	}

	return ptr;
}

void operator delete[](void* ptr) noexcept
{
	syscall(SYS_munmap, ptr, GetStackSize())
}